/*  DaemonBite (S)NES Controllers to USB Adapter
 *  Author: Mikael Norrgård <mick@daemonbite.com>
 *
 *  Copyright (c) 2020 Mikael Norrgård <http://daemonbite.com>
 *  
 *  GNU GENERAL PUBLIC LICENSE
 *  Version 3, 29 June 2007
 *  
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <https://www.gnu.org/licenses/>.
 *  
 */

#include "Gamepad.h"

// ATT: 20 chars max (including NULL at the end) according to Arduino source code.
// Additionally serial number is used to differentiate arduino projects to have different button maps!
const char *gp_serial = "NES/SNES to USB";

#define GAMEPAD_COUNT 4     // NOTE: We are going to disable CDC so more than THREE gamepads are possible

#define BUTTON_READ_DELAY 20 // Delay between button reads in µs
#define MICROS_LATCH 10      // 12µs according to specs (8 seems to work fine)
#define MICROS_CLOCK 5       //  6µs according to specs (4 seems to work fine)
#define MICROS_PAUSE 5       //  6µs according to specs (4 seems to work fine)

#define UP 0x01
#define DOWN 0x02
#define LEFT 0x04
#define RIGHT 0x08

#define NES1 0
#define NES2 1
#define SNES1 2
#define SNES2 3

// Wire it all up according to the following table:
//
// NES           SNES        Arduino Pro Micro
// --------------------------------------
// VCC                       VCC (All gamepads)
// GND                       GND (All gamepads)
// OUT0 (LATCH)              2   (PD1, All gamepads)
// CUP  (CLOCK)              3   (PD0, All gamepads)
// D1   (GP1: DATA)          A0  (PF7, Gamepad 1)
// D1   (GP2: DATA)          A1  (PF6, Gamepad 2)
// D1   (GP3: DATA)          A2  (PF5, Gamepad 3)
// D1   (GP4: DATA)          A3  (PF4, Gamepad 4)

enum ControllerType
{
    NONE,
    NES,
    SNES
};

// Set up USB HID gamepads
Gamepad_ Gamepad[GAMEPAD_COUNT];

// Controllers
uint32_t buttons[GAMEPAD_COUNT] = {0, 0, 0, 0};
uint32_t buttonsPrev[GAMEPAD_COUNT] = {0, 0, 0, 0};
uint8_t gpBit[GAMEPAD_COUNT] = {B10000000, B01000000, B00100000, B00010000};
ControllerType controllerType[GAMEPAD_COUNT] = {NES, NES, SNES, SNES};
uint32_t btnBits[12] = {0x10, 0x40, 0x400, 0x800, UP, DOWN, LEFT, RIGHT, 0x20, 0x80, 0x100, 0x200};
uint8_t gp = 0;
uint8_t buttonCountNES = 8;
uint8_t buttonCountSNES = 12;

// Timing
uint32_t microsButtons = 0;

void setup()
{
    // Setup latch and clock pins (2,3 or PD1, PD0)
    DDRD |= B00000011;   // output
    PORTD &= ~B00000011; // low

    // Setup data pins (A0-A3 or PF7-PF4)
    DDRF &= ~B11110000; // inputs
    PORTF |= B11110000; // enable internal pull-ups

    delay(500);
    detectControllerTypes();
}

void loop()
{
    while (1)
    {
        // See if enough time has passed since last button read
        if ((micros() - microsButtons) > BUTTON_READ_DELAY)
        {
            // Pulse latch
            sendLatch();

            for (uint8_t btn = 0; btn < buttonCountSNES; btn++)
            {
                if (btn < buttonCountNES)
                { // NES Buttons
                    (PINF & gpBit[NES1]) ? buttons[NES1] &= ~btnBits[btn] : buttons[NES1] |= btnBits[btn];
                    (PINF & gpBit[NES2]) ? buttons[NES2] &= ~btnBits[btn] : buttons[NES2] |= btnBits[btn];
                }
                if (btn < buttonCountSNES)
                { // SNES Buttons
                    (PINF & gpBit[SNES1]) ? buttons[SNES1] &= ~btnBits[btn] : buttons[SNES1] |= btnBits[btn];
                    (PINF & gpBit[SNES2]) ? buttons[SNES2] &= ~btnBits[btn] : buttons[SNES2] |= btnBits[btn];
                }
                sendClock();
            }

            // NES Controller 1
            bitWrite(buttons[NES1], 5, bitRead(buttons[NES1], 4));
            bitWrite(buttons[NES1], 4, bitRead(buttons[NES1], 6));
            buttons[NES1] &= 0xC3F;

            // NES Controller 2
            bitWrite(buttons[NES2], 5, bitRead(buttons[NES2], 4));
            bitWrite(buttons[NES2], 4, bitRead(buttons[NES2], 6));
            buttons[NES2] &= 0xC3F;

            // SNES Controller 1
            buttons[SNES1] &= 0xFFF;

            // SNES Controller 2
            buttons[SNES2] &= 0xFFF;

            for (gp = 0; gp < GAMEPAD_COUNT; gp++)
            {
                // Has any buttons changed state?
                if (buttons[gp] != buttonsPrev[gp])
                {
                    Gamepad[gp]._GamepadReport.buttons = (buttons[gp] >> 4); // First 4 bits are the axes
                    Gamepad[gp]._GamepadReport.Y = ((buttons[gp] & DOWN) >> 1) - (buttons[gp] & UP);
                    Gamepad[gp]._GamepadReport.X = ((buttons[gp] & RIGHT) >> 3) - ((buttons[gp] & LEFT) >> 2);
                    buttonsPrev[gp] = buttons[gp];
                    Gamepad[gp].send();
                }
            }

            microsButtons = micros();
        }
    }
}

void detectControllerTypes()
{
    controllerType[0] = NES;
    controllerType[1] = NES;
    controllerType[2] = SNES;
    controllerType[3] = SNES;
    buttonCountNES = 8;
    buttonCountSNES = 12;
}

void sendLatch()
{
    // Send a latch pulse to (S)NES controller(s)
    PORTD |= B00000010; // Set HIGH
    delayMicroseconds(MICROS_LATCH);
    PORTD &= ~B00000010; // Set LOW
    delayMicroseconds(MICROS_PAUSE);
}

void sendClock()
{
    // Send a clock pulse to (S)NES controller(s)
    PORTD |= B10000001; // Set HIGH
    delayMicroseconds(MICROS_CLOCK);
    PORTD &= ~B10000001; // Set LOW
    delayMicroseconds(MICROS_PAUSE);
}
